<?php
namespace nur\sery\db;

use nur\sery\cl;
use nur\sery\php\func;
use nur\sery\ValueException;

/**
 * Class CapacitorStorage: objet permettant d'accumuler des données pour les
 * réutiliser plus tard
 */
abstract class CapacitorStorage {
  abstract function db(): IDatabase;

    /** @var CapacitorChannel[] */
  protected $channels;

  function addChannel(CapacitorChannel $channel): CapacitorChannel {
    $this->_create($channel);
    $this->channels[$channel->getName()] = $channel;
    return $channel;
  }

  protected function getChannel(?string $name): CapacitorChannel {
    $name = CapacitorChannel::verifix_name($name);
    $channel = $this->channels[$name] ?? null;
    if ($channel === null) {
      $channel = $this->addChannel(new CapacitorChannel($name));
    }
    return $channel;
  }

  /** DOIT être défini dans les classes dérivées */
  const PRIMARY_KEY_DEFINITION = null;

  const COLUMN_DEFINITIONS = [
    "item__" => CapacitorChannel::SERIAL_DEFINITION,
    "item__sum_" => CapacitorChannel::SUM_DEFINITION,
    "created_" => "datetime",
    "modified_" => "datetime",
  ];

  protected function ColumnDefinitions(CapacitorChannel $channel): array {
    $definitions = [];
    if ($channel->getPrimaryKeys() === null) {
      $definitions[] = static::PRIMARY_KEY_DEFINITION;
    }
    $definitions[] = static::COLUMN_DEFINITIONS;
    $definitions[] = $channel->getColumnDefinitions();
    return cl::merge(...$definitions);
  }

  /** sérialiser les valeurs qui doivent l'être dans $values */
  protected function serialize(CapacitorChannel $channel, ?array $values): ?array {
    if ($values === null) return null;
    $cols = $this->ColumnDefinitions($channel);
    $index = 0;
    $row = [];
    foreach (array_keys($cols) as $col) {
      $key = $col;
      if ($key === $index) {
        $index++;
      } elseif ($channel->isSerialCol($key)) {
        [$serialCol, $sumCol] = $channel->getSumCols($key);
        if (array_key_exists($key, $values)) {
          $sum = $channel->getSum($key, $values[$key]);
          $row[$serialCol] = $sum[$serialCol];
          if (array_key_exists($sumCol, $cols)) {
            $row[$sumCol] = $sum[$sumCol];
          }
        }
      } else {
        if (array_key_exists($key, $values)) {
          $row[$col] = $values[$key];
        }
      }
    }
    return $row;
  }

  /** désérialiser les valeurs qui doivent l'être dans $values */
  protected function unserialize(CapacitorChannel $channel, ?array $row): ?array {
    if ($row === null) return null;
    $cols = $this->ColumnDefinitions($channel);
    $index = 0;
    $values = [];
    foreach (array_keys($cols) as $col) {
      $key = $col;
      if ($key === $index) {
        $index++;
      } elseif (!array_key_exists($col, $row)) {
      } elseif ($channel->isSerialCol($key)) {
        $value = $row[$col];
        if ($value !== null) $value = $channel->unserialize($value);
        $values[$key] = $value;
      } else {
        $values[$key] = $row[$col];
      }
    }
    return $values;
  }

  function getPrimaryKeys(CapacitorChannel $channel): array {
    $primaryKeys = $channel->getPrimaryKeys();
    if ($primaryKeys === null) $primaryKeys = ["id_"];
    return $primaryKeys;
  }

  function getRowIds(CapacitorChannel $channel, ?array $row, ?array &$primaryKeys=null): ?array {
    $primaryKeys = $this->getPrimaryKeys($channel);
    $rowIds = cl::select($row, $primaryKeys);
    if (cl::all_n($rowIds)) return null;
    else return $rowIds;
  }

  abstract function _exists(CapacitorChannel $channel): bool;

  /** tester si le canal spécifié existe */
  function exists(?string $channel): bool {
    return $this->_exists($this->getChannel($channel));
  }

  abstract function _ensureExists(CapacitorChannel $channel): void;

  /** s'assurer que le canal spécifié existe */
  function ensureExists(?string $channel): void {
    $this->_ensureExists($this->getChannel($channel));
  }

  abstract function _reset(CapacitorChannel $channel, bool $ensureExists=false): void;

  /** supprimer le canal spécifié */
  function reset(?string $channel, bool $ensureExists=false): void {
    $this->_reset($this->getChannel($channel), $ensureExists);
  }

  function _charge(CapacitorChannel $channel, $item, $func, ?array $args): int {
    $this->_create($channel);
    $tableName = $channel->getTableName();
    $now = date("Y-m-d H:i:s");
    $row = cl::merge(
      $channel->getSum("item", $item),
      $this->serialize($channel, $channel->getItemValues($item)));
    $prow = null;
    $rowIds = $this->getRowIds($channel, $row, $primaryKeys);
    if ($rowIds !== null) {
      # modification
      $prow = $this->db()->one([
        "select",
        "from" => $tableName,
        "where" => $rowIds,
      ]);
    }

    $insert = null;
    if ($prow === null) {
      # création
      $row = cl::merge($row, [
        "created_" => $now,
        "modified_" => $now,
      ]);
      $insert = true;
      if ($func === null) $func = "->onCreate";
      func::ensure_func($func, $channel, $args);
      $values = $this->unserialize($channel, $row);
      $args = [$item, $values, ...$args];
    } else {
      # modification
      if ($channel->_wasSumModified("item", $row, $prow)) {
        $insert = false;
        $row = cl::merge($row, [
          "modified_" => $now,
        ]);
      }
      if ($func === null) $func = "->onUpdate";
      func::ensure_func($func, $channel, $args);
      $values = $this->unserialize($channel, $row);
      $pvalues = $this->unserialize($channel, $prow);
      $args = [$item, $values, $pvalues, ...$args];
    }

    $updates = func::call($func, ...$args);
    if (is_array($updates) && $updates) {
      if ($insert === null) $insert = false;
      if (!array_key_exists("modified_", $updates)) {
        $updates["modified_"] = $now;
      }
      $row = cl::merge($row, $this->serialize($channel, $updates));
    }

    if ($insert === null) {
      # aucune modification
      return 0;
    } elseif ($insert) {
      $this->db()->exec([
        "insert",
        "into" => $tableName,
        "values" => $row,
      ]);
    } else {
      $this->db()->exec([
        "update",
        "table" => $tableName,
        "values" => $row,
        "where" => $rowIds,
      ]);
    }
    return 1;
  }

  /**
   * charger une valeur dans le canal
   *
   * Si $func!==null, après avoir calculé les valeurs des clés supplémentaires
   * avec {@link CapacitorChannel::getItemValues()}, la fonction est appelée
   * avec la signature de {@link CapacitorChannel::onCreate()} ou
   * {@link CapacitorChannel::onUpdate()} en fonction du type d'opération:
   * création ou mise à jour
   *
   * Si la fonction retourne un tableau, il est utilisé pour modifier les valeurs
   * insérées/mises à jour
   *
   * @return int 1 si l'objet a été chargé ou mis à jour, 0 s'il existait
   * déjà à l'identique dans le canal
   */
  function charge(?string $channel, $item, $func=null, ?array $args=null): int {
    return $this->_charge($this->getChannel($channel), $item, $func, $args);
  }

  function _discharge(CapacitorChannel $channel, bool $reset=true): iterable {
    $rows = $this->db()->all([
      "select item__",
      "from" => $channel->getTableName(),
    ]);
    foreach ($rows as $row) {
      yield unserialize($row['item__']);
    }
    if ($reset) $this->_reset($channel);
  }

  /** décharger les données du canal spécifié */
  function discharge(?string $channel, bool $reset=true): iterable {
    return $this->_discharge($this->getChannel($channel), $reset);
  }

  protected function verifixFilter(CapacitorChannel $channel, &$filter): void {
    if ($filter !== null && !is_array($filter)) {
      $id = $filter;
      $channel->verifixId($id);
      $filter = ["id_" => $id];
    }
    $filter = $this->serialize($channel, $filter);
  }

  function _count(CapacitorChannel $channel, $filter): int {
    $this->verifixFilter($channel, $filter);
    return $this->db()->get([
      "select count(*)",
      "from" => $channel->getTableName(),
      "where" => $filter,
    ]);
  }

  /** indiquer le nombre d'éléments du canal spécifié */
  function count(?string $channel, $filter=null): int {
    return $this->_count($this->getChannel($channel), $filter);
  }

  function _one(CapacitorChannel $channel, $filter): ?array {
    if ($filter === null) throw ValueException::null("filter");
    $this->verifixFilter($channel, $filter);
    $row = $this->db()->one([
      "select",
      "from" => $channel->getTableName(),
      "where" => $filter,
    ]);
    return $this->unserialize($channel, $row);
  }

  /**
   * obtenir la ligne correspondant au filtre sur le canal spécifié
   *
   * si $filter n'est pas un tableau, il est transformé en ["id_" => $filter]
   */
  function one(?string $channel, $filter): ?array {
    return $this->_one($this->getChannel($channel), $filter);
  }

  function _all(CapacitorChannel $channel, $filter): iterable {
    $this->verifixFilter($channel, $filter);
    $rows = $this->db()->all([
      "select",
      "from" => $channel->getTableName(),
      "where" => $filter,
    ], null, $this->getPrimaryKeys($channel));
    foreach ($rows as $key => $row) {
      yield $key => $this->unserialize($channel, $row);
    }
  }

  /**
   * obtenir les lignes correspondant au filtre sur le canal spécifié
   *
   * si $filter n'est pas un tableau, il est transformé en ["id_" => $filter]
   */
  function all(?string $channel, $filter): iterable {
    return $this->_all($this->getChannel($channel), $filter);
  }

  function _each(CapacitorChannel $channel, $filter, $func, ?array $args): int {
    if ($func === null) $func = "->onEach";
    func::ensure_func($func, $channel, $args);
    $onEach = func::_prepare($func);
    $db = $this->db();
    $tableName = $channel->getTableName();
    $commited = false;
    $count = 0;
    $db->beginTransaction();
    $commitThreshold = $channel->getEachCommitThreshold();
    try {
      $args ??= [];
      foreach ($this->_all($channel, $filter) as $rowValues) {
        $rowIds = $this->getRowIds($channel, $rowValues);
        $updates = func::_call($onEach, [$rowValues["item"], $rowValues, ...$args]);
        if (is_array($updates) && $updates) {
          if (!array_key_exists("modified_", $updates)) {
            $updates["modified_"] = date("Y-m-d H:i:s");
          }
          $db->exec([
            "update",
            "table" => $tableName,
            "values" => $this->serialize($channel, $updates),
            "where" => $rowIds,
          ]);
          if ($commitThreshold !== null) {
            $commitThreshold--;
            if ($commitThreshold == 0) {
              $db->commit();
              $db->beginTransaction();
              $commitThreshold = $channel->getEachCommitThreshold();
            }
          }
        }
        $count++;
      }
      $db->commit();
      $commited = true;
      return $count;
    } finally {
      if (!$commited) $db->rollback();
    }
  }

  /**
   * appeler une fonction pour chaque élément du canal spécifié.
   *
   * $filter permet de filtrer parmi les élements chargés
   *
   * $func est appelé avec la signature de {@link CapacitorChannel::onEach()}
   * si la fonction retourne un tableau, il est utilisé pour mettre à jour la
   * ligne
   *
   * @return int le nombre de lignes parcourues
   */
  function each(?string $channel, $filter, $func=null, ?array $args=null): int {
    return $this->_each($this->getChannel($channel), $filter, $func, $args);
  }

  function _delete(CapacitorChannel $channel, $filter, $func, ?array $args): int {
    if ($func === null) $func = "->onDelete";
    func::ensure_func($func, $channel, $args);
    $onEach = func::_prepare($func);
    $db = $this->db();
    $tableName = $channel->getTableName();
    $commited = false;
    $count = 0;
    $db->beginTransaction();
    $commitThreshold = $channel->getEachCommitThreshold();
    try {
      $args ??= [];
      foreach ($this->_all($channel, $filter) as $rowValues) {
        $rowIds = $this->getRowIds($channel, $rowValues);
        $delete = boolval(func::_call($onEach, [$rowValues["item"], $rowValues, ...$args]));
        if ($delete) {
          $db->exec([
            "delete",
            "from" => $tableName,
            "where" => $rowIds,
          ]);
          if ($commitThreshold !== null) {
            $commitThreshold--;
            if ($commitThreshold == 0) {
              $db->commit();
              $db->beginTransaction();
              $commitThreshold = $channel->getEachCommitThreshold();
            }
          }
        }
        $count++;
      }
      $db->commit();
      $commited = true;
      return $count;
    } finally {
      if (!$commited) $db->rollback();
    }
  }

  /**
   * supprimer tous les éléments correspondant au filtre et pour lesquels la
   * fonction retourne une valeur vraie si elle est spécifiée
   *
   * $filter permet de filtrer parmi les élements chargés
   *
   * $func est appelé avec la signature de {@link CapacitorChannel::onDelete()}
   * si la fonction retourne un tableau, il est utilisé pour mettre à jour la
   * ligne
   *
   * @return int le nombre de lignes parcourues
   */
  function delete(?string $channel, $filter, $func=null, ?array $args=null): int {
    return $this->_delete($this->getChannel($channel), $filter, $func, $args);
  }

  abstract function close(): void;







}
